.installed-overlay-box {
	text-shadow: none;
}

.category-tile:not(.category-tile-iconless) {
	box-shadow: inset 0 0 0 1px alpha(@card_fg_color, .5);
}

review-bar,
.review-histogram star-image {
	color: alpha(@theme_fg_color, .8);
}

star-image > image.starred {
	color: inherit;
}

star-image > image.non-starred {
	opacity: 0.4;
}

.origin-rounded-box {
	box-shadow: inset 0 0 0 1px currentColor;
}

.context-tile-lozenge {
	box-shadow: inset 0 0 0 1px currentColor;
}

app-context-bar .context-tile {
	border-color: @borders;
	box-shadow: none;
}
